package com.hortonworks.spark.sql.hive.llap;

import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import com.hortonworks.spark.sql.hive.llap.util.JobUtil;
import com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataSourceWriter;
import java.io.IOException;
import java.util.Map;
import java.util.Optional;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.Path;
import org.apache.spark.sql.SaveMode;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.sql.sources.v2.DataSourceOptions;
import org.apache.spark.sql.sources.v2.DataSourceV2;
import org.apache.spark.sql.sources.v2.ReadSupport;
import org.apache.spark.sql.sources.v2.SessionConfigSupport;
import org.apache.spark.sql.sources.v2.WriteSupport;
import org.apache.spark.sql.sources.v2.reader.DataSourceReader;
import org.apache.spark.sql.sources.v2.writer.DataSourceWriter;
import org.apache.spark.sql.types.StructType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/*
 * Driver:
 *   UserCode -> HiveWarehouseConnector -> HiveWarehouseDataSourceReader -> HiveWarehouseInputPartition
 * Task serializer:
 *   HiveWarehouseInputPartition (Driver) -> bytes -> HiveWarehouseInputPartition (Executor task)
 * Executor:
 *   HiveWarehouseInputPartition -> HiveWarehouseInputPartitionReader
 */
public class HiveWarehouseConnector implements DataSourceV2, ReadSupport, SessionConfigSupport, WriteSupport {

  private static Logger LOG = LoggerFactory.getLogger(HiveWarehouseConnector.class);

  @Override
  public DataSourceReader createReader(DataSourceOptions options) {
    try {
      return getDataSourceReader(getOptions(options));
    } catch (IOException e) {
      LOG.error("Error creating {}", getClass().getName());
      LOG.error(ExceptionUtils.getStackTrace(e));
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  @Override
  public Optional<DataSourceWriter> createWriter(String jobId, StructType schema,
      SaveMode mode, DataSourceOptions options) {
    Map<String, String> params = getOptions(options);
    String defaultStagingDirLocation = HWConf.LOAD_STAGING_DIR.getFromOptionsMap(params);
    String stagingDirPrefix = JobUtil.getStagingDirPrefix(defaultStagingDirLocation);
    Path path = new Path(stagingDirPrefix);
    Configuration conf = SparkSession.getActiveSession().get().sparkContext().hadoopConfiguration();
    JobUtil.populateSparkConfs(conf);
    return Optional.of(getDataSourceWriter(jobId, schema, path, params, conf, mode));
  }

  @Override public String keyPrefix() {
    return HiveWarehouseSession.HIVE_WAREHOUSE_POSTFIX;
  }

  private static Map<String, String> getOptions(DataSourceOptions option) {
    Map<String, String> options = option.asMap();
    if(!options.containsKey(HWConf.RESOLVED_HS2_URL.getSimpleKey()))
    {
      SparkSession spark = SparkSession.getActiveSession().get();
      //Adding Resolved jdbc url to option as  HWC session is not initialized.
      String resolvedJdbcUrl = com.hortonworks.hwc.HiveWarehouseSession.session(spark).resolveJdbcUrl();
      options.put(HWConf.RESOLVED_HS2_URL.getSimpleKey(), resolvedJdbcUrl);
      LOG.info("HWC session not initialized. Resolved jdbc url added to options = "+ resolvedJdbcUrl );
    }
    return options;
  }

  protected DataSourceReader getDataSourceReader(Map<String, String> params) throws IOException {
    return new JdbcDataSourceReader(params);
  }

  protected DataSourceWriter getDataSourceWriter(String jobId, StructType schema,
                                                 Path path, Map<String, String> options, Configuration conf, SaveMode mode) {
    return new HiveWarehouseDataSourceWriter(options, jobId, schema, path, conf, mode);
  }
}
