/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hortonworks.spark.sql.hive.llap;

import com.google.common.base.Preconditions;
import com.hortonworks.hwc.MergeBuilder;
import com.hortonworks.spark.sql.hive.llap.common.DriverResultSet;
import com.hortonworks.spark.sql.hive.llap.common.HWConf;
import com.hortonworks.spark.sql.hive.llap.common.HWConf.HWCReadMode;
import com.hortonworks.spark.sql.hive.llap.query.builder.CreateTableBuilder;
import com.hortonworks.spark.sql.hive.llap.query.builder.MergeBuilderImpl;
import com.hortonworks.spark.sql.hive.llap.util.FunctionWith4Args;
import com.hortonworks.spark.sql.hive.llap.util.HWCOptions;
import com.hortonworks.spark.sql.hive.llap.util.HiveQlUtil;
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil;
import com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod;
import com.hortonworks.spark.sql.hive.llap.util.StreamingMetaCleaner;
import com.hortonworks.spark.sql.hive.llap.util.TriFunction;
import com.qubole.spark.hiveacid.transaction.HiveAcidTxnManagerObject;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.spark.SparkConf;
import org.apache.spark.sql.DataFrameReader;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.RuntimeConfig;
import org.apache.spark.sql.SparkSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.UUID;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Supplier;
import java.util.regex.Pattern;

import static com.hortonworks.spark.sql.hive.llap.common.HWConf.DEFAULT_DB;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_HIVE_JDBC;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.ExecutionMethod.EXECUTE_JDBC_CLUSTER;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.isDataFetchQuery;
import static com.hortonworks.spark.sql.hive.llap.util.QueryExecutionUtil.resolveExecutionMethod;
import static com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat.ORC;
import static java.lang.String.format;

public class HiveWarehouseSessionImpl extends com.hortonworks.hwc.HiveWarehouseSession {

  private static final Logger LOG = LoggerFactory.getLogger(HiveWarehouseSessionImpl.class);

  private static final String SELECT_STMT_REGEX = "(\\s+)?select(\\s+).+";
  private static final Pattern selectPattern = Pattern.compile(SELECT_STMT_REGEX, Pattern.CASE_INSENSITIVE);
  public static final String SESSION_CLOSED_MSG = "Session is closed!!! No more operations can be performed. Please create a new HiveWarehouseSession";
  public static final String HWC_SESSION_ID_KEY = "hwc_session_id";
  public static final String HWC_QUERY_MODE = "hwc_query_mode";
  static String HIVE_WAREHOUSE_CONNECTOR_INTERNAL = HiveWarehouseSession.HIVE_WAREHOUSE_CONNECTOR;

  protected HiveWarehouseSessionState sessionState;

  protected Supplier<Connection> getConnector;

  protected TriFunction<Connection, String, String, DriverResultSet> executeStmt;

  protected TriFunction<Connection, String, String, Boolean> executeUpdate;

  protected FunctionWith4Args<Connection, String, String, Boolean, Boolean> executeUpdateWithPropagateException;

  private final String sessionId;
  private final AtomicReference<HwcSessionState> hwcSessionStateRef;

  private boolean usingDirectReaderMode() {
    return (HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V1) ||
            HWCReadMode.resolveReadMode(HWCReadMode.DIRECT_READER_V2) ||
            HWConf.getHwcExecutionMode().equalsIgnoreCase(HWConf.HWC_EXECUTION_MODE_SPARK));
  }

  public HiveWarehouseSessionImpl(HiveWarehouseSessionState sessionState) {
    this.sessionState = sessionState;
    this.sessionId = UUID.randomUUID().toString();
    getConnector = () -> DefaultJDBCWrapper.getConnector(sessionState);
    executeStmt = (conn, database, sql) ->
            DefaultJDBCWrapper.executeStmt(conn, database, sql, 0);
    executeUpdate = (conn, database, sql) ->
        DefaultJDBCWrapper.executeUpdate(conn, database, sql);
    executeUpdateWithPropagateException = DefaultJDBCWrapper::executeUpdate;
    hwcSessionStateRef = new AtomicReference<>(HwcSessionState.OPEN);

    LOG.info("Created a new HWC session: {}", sessionId);
  }

  private enum HwcSessionState {
    OPEN, CLOSED;
  }

  private void ensureSessionOpen() {
    Preconditions.checkState(HwcSessionState.OPEN.equals(hwcSessionStateRef.get()), SESSION_CLOSED_MSG);
  }

  public Dataset<Row> q(String sql) {
    ensureSessionOpen();
    return executeQuery(sql);
  }

  @Override
  public Dataset<Row> sql(String queryToFetchData) {

    if (usingDirectReaderMode()) {
      //datasource v1 will handle both old and new direct-reader
      return executeQueryInternal(queryToFetchData, null);
    } else if (HWCReadMode.resolveReadMode(HWCReadMode.JDBC_CLIENT)) {
      return executeSimpleJdbc(queryToFetchData);
    }
    else if(HWCReadMode.resolveReadMode(HWCReadMode.SECURE_ACCESS)
            && isDataFetchQuery(queryToFetchData)){
      return executeQuerySecureAccess(queryToFetchData);
    }
    else {
      // jdbc cluster
      return executeJdbcInternal(queryToFetchData);
    }
  }

  public Dataset<Row> executeQuery(String sql) {
    return executeSmart(EXECUTE_JDBC_CLUSTER, sql, null);
  }

  public Dataset<Row> executeQuery(String sql, boolean useSplitsEqualToSparkCores) {
    return executeSmart(EXECUTE_JDBC_CLUSTER, sql, getCoresInSparkCluster());
  }

  public Dataset<Row> executeQuery(String sql, int numSplitsToDemand) {
    return executeSmart(EXECUTE_JDBC_CLUSTER, sql, numSplitsToDemand);
  }

  private Dataset<Row> executeQueryInternal(String sql, Integer numSplitsToDemand) {
    // If query is select query, then used HiveWarehouseDataSource or else use Simple JDBC in case of DDL operations.
    Preconditions.checkNotNull(sql,"The query cannot be null");
    if (isDataFetchQuery(sql)) {
      ensureSessionOpen();

      if(HWCReadMode.resolveReadMode(HWCReadMode.SECURE_ACCESS)){
        return executeQuerySecureAccess(sql);
      }

      else if (usingDirectReaderMode()) {
        if (!HWConf.getSparkSqlExtension(sessionState).
                contains("com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension")) {
          LOG.info("For spark execution, set " +
                  "spark.sql.extensions to com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension");
          LOG.warn(" com.qubole.spark.hiveacid.HiveAcidAutoConvertExtension will be replaced by " +
                  "com.hortonworks.spark.sql.rule.Extensions." +
                  "Please switch to com.hortonworks.spark.sql.rule.Extensions.");
        }
        return session().sql(sql);
      }
      return executeJdbcInternal(sql);
    } else {
      return executeSimpleJdbc(sql);
    }
  }

  private String setSplitPropertiesQuery(int numSplitsToDemand) {
    return String.format("SET tez.grouping.split-count=%s", numSplitsToDemand);
  }

  private int getCoresInSparkCluster() {
    //this was giving good results(number of cores) when tested in standalone and yarn mode(without dynamicAllocation)
    return session().sparkContext().defaultParallelism();
  }

  public Dataset<Row> execute(String sql) {
    return executeSmart(EXECUTE_HIVE_JDBC, sql, null);
  }

  private Dataset<Row> executeSmart(ExecutionMethod current, String sql, Integer numSplitsToDemand) {
    ExecutionMethod resolved = resolveExecutionMethod(
        BooleanUtils.toBoolean(HWConf.SMART_EXECUTION.getString(sessionState)), current, sql);
    if (EXECUTE_HIVE_JDBC.equals(resolved)) {
      return executeSimpleJdbc(sql);
    }
    return executeQueryInternal(sql, numSplitsToDemand);
  }

  public String getJdbcUrlForExecutor() {
    RuntimeConfig conf = session().conf();
    if ((conf.contains(HWConf.HIVESERVER2_CREDENTIAL_ENABLED) && conf.get(HWConf.HIVESERVER2_CREDENTIAL_ENABLED).equals("true"))
            || conf.contains(HWConf.HIVESERVER2_JDBC_URL_PRINCIPAL)) {
      String master = conf.get(HWConf.SPARK_MASTER, "local");
      LOG.debug("Getting jdbc connection url for kerberized cluster with spark.master = {}", master);
      if (master.equals("yarn")) {
        StringBuilder jdbcURLBuilder = new StringBuilder();
        // In YARN mode for kerberized clusters, executors always run on yarn application. So, need
        // to use delegationToken to authenticate HS2 connection from it.
        jdbcURLBuilder.append(format("%s;auth=delegationToken", conf.get(HWConf.HIVESERVER2_JDBC_URL)));
        if (conf.contains(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST) &&
                StringUtils.isNotBlank(conf.get(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST))) {
          jdbcURLBuilder.append("?").append(conf.get(HWConf.HIVESERVER2_JDBC_URL_CONF_LIST));
        }
        return jdbcURLBuilder.toString();
      }
    }
    return HWConf.RESOLVED_HS2_URL.getString(sessionState);
  }

  private Dataset<Row> executeJdbcInternal(String sql) {
    // If query is select query, then used JDBCDataSource or else use Simple JDBC in case of DDL operations.
    Preconditions.checkNotNull(sql,"The query cannot be null");
    if (isDataFetchQuery(sql)) {
      ensureSessionOpen();
      String mode = EXECUTE_HIVE_JDBC.name();
      DataFrameReader dfr = session().read().format(HIVE_WAREHOUSE_CONNECTOR_INTERNAL).option("query", sql)
              .option(HWC_SESSION_ID_KEY, sessionId)
              .option(HWC_QUERY_MODE, mode)
              .option(QueryExecutionUtil.HIVE_JDBC_URL_FOR_EXECUTOR, getJdbcUrlForExecutor());
      return dfr.load();
    } else {
      return executeSimpleJdbc(sql);
    }
  }

  private Dataset<Row> executeQuerySecureAccess(final String sql) {
    LOG.info("execution mode :Secure Access FGAC");
    Preconditions.checkNotNull(sql, "The query cannot be null");
    ensureSessionOpen();
    return SecureAccessModeExecutor.getInstance().execute(sessionState,sql);
  }


  public boolean executeUpdate(String sql) {
    return executeUpdate(sql, true);
  }

  @Override
  public boolean executeUpdate(String sql, boolean propagateException) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      return executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState), sql, propagateException);
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  private boolean executeUpdateInternal(String sql, Connection conn) {
    ensureSessionOpen();
    return executeUpdateWithPropagateException.apply(conn, DEFAULT_DB.getString(sessionState), sql, true);
  }

  public Dataset<Row> table(String table) {
    ensureSessionOpen();
    String mode = EXECUTE_JDBC_CLUSTER.name();
    DataFrameReader dfr = session().read().format(HIVE_WAREHOUSE_CONNECTOR_INTERNAL).option("table", table)
        .option(HWC_SESSION_ID_KEY, sessionId).option(HWC_QUERY_MODE, mode)
        .option(QueryExecutionUtil.HIVE_JDBC_URL_FOR_EXECUTOR, getJdbcUrlForExecutor());
    return dfr.load();
  }

  public SparkSession session() {
    ensureSessionOpen();
    return sessionState.session;
  }

  // Exposed for Python side.
  public HiveWarehouseSessionState sessionState() {
    return sessionState;
  }

  SparkConf conf() {
    return sessionState.session.sparkContext().getConf();
  }

  /* Catalog helpers */
  public void setDatabase(String name) {
    ensureSessionOpen();
    HWConf.DEFAULT_DB.setString(sessionState, name);
  }

  private Dataset<Row> executeSimpleJdbc(String sql) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      DriverResultSet drs = executeStmt.apply(conn, DEFAULT_DB.getString(sessionState), sql);
      return drs.asDataFrame(session());
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  public Dataset<Row> showDatabases() {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.showDatabases());
  }

  public Dataset<Row> showTables() {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.showTables(DEFAULT_DB.getString(sessionState)));
  }

  public Dataset<Row> describeTable(String table) {
    ensureSessionOpen();
    return executeSimpleJdbc(HiveQlUtil.describeTable(DEFAULT_DB.getString(sessionState), table));
  }

  public void dropDatabase(String database, boolean ifExists, boolean cascade) {
    executeUpdate(HiveQlUtil.dropDatabase(database, ifExists, cascade));
  }

  public void dropTable(String table, boolean ifExists, boolean purge) {
    ensureSessionOpen();
    try (Connection conn = getConnector.get()) {
      executeUpdateInternal(HiveQlUtil.useDatabase(DEFAULT_DB.getString(sessionState)), conn);
      String dropTable = HiveQlUtil.dropTable(table, ifExists, purge);
      executeUpdateInternal(dropTable, conn);
    } catch (SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  public void createDatabase(String database) {
    executeUpdate(HiveQlUtil.createDatabase(database, false), true);
  }


  public void createDatabase(String database, boolean ifNotExists) {
    executeUpdate(HiveQlUtil.createDatabase(database, ifNotExists), true);
  }

  public CreateTableBuilder createTable(String tableName) {
    ensureSessionOpen();
    return new CreateTableBuilder(this, DEFAULT_DB.getString(sessionState), tableName, ORC, new HWCOptions(new HashMap<>()));
  }

  @Override
  public MergeBuilder mergeBuilder() {
    ensureSessionOpen();
    return new MergeBuilderImpl(this, DEFAULT_DB.getString(sessionState));
  }

  @Override
  public void cleanUpStreamingMeta(String queryCheckpointDir, String dbName, String tableName) {
    ensureSessionOpen();
    try {
      new StreamingMetaCleaner(sessionState, queryCheckpointDir, dbName, tableName).clean();
    } catch (IOException | SQLException e) {
      throw new RuntimeException(e.getMessage(), e);
    }
  }

  @Override
  public void commitTxn() {
    if (usingDirectReaderMode()) {
      HiveAcidTxnManagerObject.commitTxn(session());
    }
  }

  @Override
  public void close() {
    try {
      commitTxn();
    } catch (Exception e) {
      LOG.info("Failed to end hive-acid txn. The txn may be committed already.", e);
    }
    Preconditions.checkState(hwcSessionStateRef.compareAndSet(HwcSessionState.OPEN, HwcSessionState.CLOSED),
        SESSION_CLOSED_MSG);
  }

  public String getSessionId() {
    return this.sessionId;
  }

  protected void finalize() {
    close();
  }
}

