/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.hortonworks.spark.sql.hive.llap.query.builder;

import com.hortonworks.spark.sql.hive.llap.HiveWarehouseSession;
import com.hortonworks.spark.sql.hive.llap.util.HWCOptions;
import com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import org.apache.commons.lang3.tuple.Pair;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static com.hortonworks.spark.sql.hive.llap.util.HWCOptions.*;
import static com.hortonworks.spark.sql.hive.llap.util.HiveQlUtil.*;
import static com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat.ORC;
import static com.hortonworks.spark.sql.hive.llap.writers.HiveWarehouseDataWriterHelper.FileFormat.TEXT;
import static java.lang.String.join;

public class CreateTableBuilder implements com.hortonworks.hwc.CreateTableBuilder {
    private final HiveWarehouseSession hive;
    private String database;
    private final String tableName;
    private FileFormat fileFormat;
    private HWCOptions hwcOptions;
    private boolean ifNotExists;
    private List<Pair<String, String>> cols = new ArrayList<>();
    private List<Pair<String, String>> parts = new ArrayList<>();
    private Map<String, String> options = new HashMap<>();
    private List<Pair<String, String>> props = new ArrayList<>();
    private String[] clusters;
    private Long buckets;
    private boolean propagateException;
    private static Logger LOG = LoggerFactory.getLogger(CreateTableBuilder.class);

    public CreateTableBuilder(HiveWarehouseSession hive, String database, String tableName, FileFormat fileFormat, HWCOptions hwcOptions) {
        this.hive = hive;
        this.tableName = tableName;
        this.fileFormat = fileFormat;
        this.database = database;
        this.hwcOptions = hwcOptions;
    }

    @Override
    public CreateTableBuilder ifNotExists() {
        this.ifNotExists = true;
        return this;
    }

    @Override
    public CreateTableBuilder propagateException() {
      this.propagateException = true;
      return this;
    }

    @Override
    public CreateTableBuilder database(String database) {
        this.database = database;
        return this;
    }

    @Override
    public CreateTableBuilder fileFormat(String fileFormat) {
        this.fileFormat = FileFormat.getFormat(fileFormat);
        return this;
    }

    @Override
    public CreateTableBuilder column(String name, String type) {
        cols.add(Pair.of(name, type));
        return this;
    }

    @Override
    public CreateTableBuilder partition(String name, String type) {
        parts.add(Pair.of(name, type));
        return this;
    }

    @Override
    public CreateTableBuilder prop(String key, String value) {
        props.add(Pair.of(key, value));
        return this;
    }

    @Override
    public CreateTableBuilder option(String key, String value) {
        options.put(key,value);
        return this;
    }

    @Override
    public CreateTableBuilder clusterBy(long numBuckets, String ... columns) {
        this.buckets = numBuckets;
        this.clusters = columns;
        return this;
    }

    @Override
    public void create() {
        hwcOptions = new HWCOptions(options);
        hive.executeUpdate(this.toString());
    }

    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append(createTablePrelude(database, tableName, ifNotExists));
        if(cols.size() > 0) {
            List<String> colsStrings = new ArrayList<>();
            for(Pair<String, String> col : cols) {
                colsStrings.add(col.getKey() + " " + col.getValue());
            }
            builder.append(columnSpec(join(",", colsStrings)));
        }
        if(parts.size() > 0) {
            List<String> partsStrings = new ArrayList<>();
            for(Pair<String, String> part : parts) {
                partsStrings.add(part.getKey() + " " + part.getValue());
            }
            builder.append(partitionSpec(join(",", partsStrings)));
        }
        if(clusters != null) {
            builder.append(bucketSpec(join(",", clusters), buckets));
        }
        if (fileFormat.equals(TEXT)) {
            builder.append(getRowFormatString(hwcOptions));
        }
        builder.append(" STORED AS " + fileFormat.toString() + " ");
        List<Pair<String, String>> tableProps = new ArrayList<>();
        tableProps.add(Pair.of(TRANSACTIONAL, hwcOptions.transactional));
        switch (hwcOptions.transactional_properties) {
            case NONE:
                if (fileFormat.equals(ORC)) break;
            case INSERT_ONLY:
                tableProps.add(Pair.of(TRANSACTIONAL_PROPERTIES, INSERT_ONLY));
                break;
            case DEFAULT:
                if (fileFormat.equals(ORC)) break;
                throw new IllegalArgumentException("Transactional property " + DEFAULT + " can't be set for " + fileFormat + " format. ");
            default:
                throw new IllegalArgumentException("Invalid transactional property " + hwcOptions.transactional_properties +
                        " specified for " + tableName);
        }
        builder.append(getTablePropertiesString(props, tableProps));
        String builderString = builder.toString();
        LOG.debug("Builder String : {}", builderString);
        return builderString;
    }
}
