/*
 * Copyright 2019 Qubole, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid

import com.qubole.spark.hiveacid.hive.HiveAcidMetadata
import com.qubole.spark.hiveacid.reader.TableReader
import com.qubole.spark.hiveacid.reader.hive.HiveAcidSearchArgument
import com.qubole.spark.hiveacid.transaction.HiveAcidTxn
import java.util.OptionalLong
import org.apache.hadoop.hive.metastore.HiveMetaStoreClient
import org.apache.spark.SparkContext
import org.apache.spark.internal.Logging
import org.apache.spark.sql.SparkSession
import org.apache.spark.sql.sources.Filter
import org.apache.spark.sql.sources.v2.DataSourceV2
import org.apache.spark.sql.sources.v2.reader.{DataSourceReader, InputPartition, Statistics, SupportsPushDownFilters, SupportsPushDownRequiredColumns, SupportsReportStatistics, SupportsScanColumnarBatch}
import org.apache.spark.sql.types.StructType
import org.apache.spark.sql.vectorized.ColumnarBatch


class HiveTableStatisticsForV2 (rows: OptionalLong,
                            physicalSize /*In bytes*/ : OptionalLong)
                            extends Statistics {
  override def sizeInBytes(): OptionalLong = rows
  override def numRows(): OptionalLong = physicalSize
}

/**
  * Data source V2 implementation for HiveACID
*/
class HiveAcidDataSourceV2Reader
  extends DataSourceV2 with DataSourceReader with SupportsScanColumnarBatch
    with SupportsPushDownRequiredColumns
    with SupportsPushDownFilters
    with SupportsReportStatistics
    with Logging {

  def this(options: java.util.Map[String, String],
           sparkSession : SparkSession,
           dbName : String,
           tblName : String) {
    this()
    this.options = options
    this.sparkSession = sparkSession
    val parameters = scala.collection.immutable.Map[String, String] ()
    this.sparkConf = SparkAcidConf (sparkSession, parameters)

    if (dbName != null) {
      hiveAcidMetadata = HiveAcidMetadata.fromSparkSession(sparkSession, dbName + "." + tblName)
    } else {
      hiveAcidMetadata = HiveAcidMetadata.fromSparkSession(sparkSession, tblName)
    }
    //this is a hack to prevent the following situation:
    //Spark(v 2.4.0) creates one instance of DataSourceReader to call readSchema() and then a new instance of DataSourceReader
    //to call pushFilters(), planBatchInputPartitions() etc. Since it uses different DataSourceReader instances,
    //and reads schema in former instance, schema remains null in the latter instance(which causes problems for other methods).
    //More discussion: http://apache-spark-user-list.1001560.n3.nabble.com/DataSourceV2-APIs-creating-multiple-instances-of-DataSourceReader-and-hence-not-preserving-the-state-tc33646.html
    //Also a null check on schema is already there in readSchema() to prevent initialization more than once just in case.
    readSchema
    //sessionId = getCurrentSessionId
  }

  private var options: java.util.Map[String, String] = null
  private var sparkSession : SparkSession = null
  private var sparkConf : SparkAcidConf = null

  private var curTxn: HiveAcidTxn = _

  //The pruned schema
  private var schema: StructType = null

  private var pushedFilterArray : Array[Filter] = null

  private var hiveAcidMetadata: HiveAcidMetadata = _

  private lazy val client: HiveMetaStoreClient =
    new HiveMetaStoreClient(SparkContext.getOrCreate.hadoopConfiguration, null, false)

  override def readSchema: StructType = {
    if (schema == null) {
      schema = hiveAcidMetadata.tableSchema
    }
    schema
  }

  override def planBatchInputPartitions() : java.util.List[InputPartition[ColumnarBatch]] = {
    val factories = new java.util.ArrayList[InputPartition[ColumnarBatch]]
    inTxn {
      val reader = new TableReader(sparkSession, curTxn, hiveAcidMetadata)
      val hiveReader = reader.getReader(schema.fieldNames,
        pushedFilterArray, sparkConf)
      factories.addAll(hiveReader)
    }
    factories
  }

  def endTxn(): Unit = if (curTxn != null && !curTxn.istxnClosed()) {
    curTxn.end()
    curTxn = null
  }

  // Start local transaction if not passed.
  private def getOrCreateTxn(): Unit = {
    // The same relation can be used to create multiple rdds. Each of the rdd will have same txn snapshot.
    if (curTxn != null) {
      logInfo(s"active txn found for hive table $curTxn")
      return
    }

    curTxn = HiveAcidTxn.createTransaction(sparkSession)
    curTxn.begin(false)
    logDebug(s"Started normal Transactions $curTxn")
  }

  // Start and end transaction under protection.
  // The transaction opened here will be closed by SparkAcidQueryListener when the query finishes running.
  private def inTxn(f: => Unit): Unit = synchronized {
    getOrCreateTxn()
    var abort = false
    try { f }
    catch {
      case e: Exception =>
        logError("Unable to execute in transactions due to: " + e.getMessage)
        abort = true;
    }
  }

  override def pushFilters (filters: Array[Filter]): Array[Filter] = {
    this.pushedFilterArray = HiveAcidSearchArgument.
      getSupportedFilters(hiveAcidMetadata.tableSchema, filters.toSeq).toArray
    // ORC does not do row level filtering. So the filters has to be applied again.
    filters
  }

  override def pushedFilters(): Array[Filter] = this.pushedFilterArray

  override def pruneColumns(requiredSchema: StructType): Unit = {
    this.schema = requiredSchema
  }

  override def estimateStatistics(): Statistics = {
    new HiveTableStatisticsForV2 (
    OptionalLong.of (hiveAcidMetadata.hTable.getParameters
      .getOrDefault("totalSize", sparkSession.sessionState.conf.defaultSizeInBytes.toString).toLong),
    OptionalLong.empty)
  }
}
