/*
 * Copyright 2019 Qubole, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid.reader.hive

import com.qubole.spark.hiveacid.{HiveAcidErrors, SparkAcidConf}
import com.qubole.spark.hiveacid.hive.{HiveAcidMetadata, HiveConverter}
import com.qubole.spark.hiveacid.rdd.{EmptyRDD, HiveAcidPartition, HiveAcidRDD, HiveAcidUnionRDD}
import com.qubole.spark.hiveacid.reader.v2.HiveAcidInputPartitionV2
import com.qubole.spark.hiveacid.reader.{Reader, ReaderOptions, ReaderPartition}
import com.qubole.spark.hiveacid.util.{HWCDateTimeUtils, HiveAcidCommon, HiveAcidInputPartitionGeneratorV2, HiveSplitInfoV2, SerializableConfiguration, Util}
import org.apache.spark.sql.catalyst.expressions.{Attribute, Literal, SpecificInternalRow}
import org.apache.hadoop.conf.Configuration
import org.apache.hadoop.fs.{Path, PathFilter}
import org.apache.hadoop.hive.common.{ValidTxnList, ValidWriteIdList}
import org.apache.hadoop.hive.conf.HiveConf.ConfVars
import org.apache.hadoop.hive.metastore.api.FieldSchema
import org.apache.hadoop.hive.metastore.api.hive_metastoreConstants.META_TABLE_STORAGE
import org.apache.hadoop.hive.metastore.utils.MetaStoreUtils.{getColumnNamesFromFieldSchema, getColumnTypesFromFieldSchema}
import org.apache.hadoop.hive.ql.exec.Utilities
import org.apache.hadoop.hive.ql.io.{AcidUtils, RecordIdentifier}
import org.apache.hadoop.hive.ql.io.sarg.ConvertAstToSearchArg
import org.apache.hadoop.hive.ql.metadata.{Partition => HiveJarPartition, Table => HiveTable}
import org.apache.hadoop.hive.ql.plan.TableDesc
import org.apache.hadoop.hive.serde2.{ColumnProjectionUtils, Deserializer}
import org.apache.hadoop.hive.serde2.objectinspector.primitive.{BinaryObjectInspector, BooleanObjectInspector, ByteObjectInspector, DateObjectInspector, DoubleObjectInspector, FloatObjectInspector, HiveCharObjectInspector, HiveDecimalObjectInspector, HiveVarcharObjectInspector, IntObjectInspector, LongObjectInspector, ShortObjectInspector, TimestampObjectInspector}
import org.apache.hadoop.hive.serde2.objectinspector.{ObjectInspectorConverters, StructObjectInspector}
import org.apache.hadoop.io.Writable
import org.apache.hadoop.mapred.{FileInputFormat, InputFormat, JobConf}
import org.apache.spark.broadcast.Broadcast
import org.apache.spark.deploy.SparkHadoopUtil
import org.apache.spark.internal.Logging
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.catalyst.InternalRow
import org.apache.spark.sql.catalyst.analysis.CastSupport
import org.apache.spark.sql.catalyst.catalog.CatalogTablePartition
import org.apache.spark.sql.catalyst.util.DateTimeUtils
import org.apache.spark.sql.hive.{Hive3Inspectors, HiveAcidUtils}
import org.apache.spark.sql.internal.SQLConf
import org.apache.spark.sql.sources.Filter
import org.apache.spark.sql.sources.v2.reader.InputPartition
import org.apache.spark.sql.types.{StructField, StructType}
import org.apache.spark.sql.vectorized.ColumnarBatch
import org.apache.spark.sql.{SparkSession, functions}
import org.apache.spark.unsafe.types.UTF8String
import java.util
import java.util.Properties
import scala.collection.JavaConverters.{asScalaBufferConverter, propertiesAsScalaMapConverter}

/**
 * Helper class for scanning tables stored in Hadoop - e.g., to read
 * Hive tables that reside in the data warehouse directory.
 * @param sparkSession - spark session
 * @param readerOptions - reader options for creating RDD
 * @param hiveAcidOptions - hive related reader options for creating RDD
 * @param validWriteIds - validWriteIds
 * @param validTxnList - The valid txn list for the current snapshot
 * @param partitions - The list of partitions to be scanned. Valid only for partitioned table.
 */
private[reader] class HiveAcidReader(sparkSession: SparkSession,
                                     readerOptions: ReaderOptions,
                                     hiveAcidOptions: HiveAcidReaderOptions,
                                     validWriteIds: ValidWriteIdList,
                                     validTxnList : ValidTxnList,
                                     partitions: Seq[ReaderPartition])

extends CastSupport with Reader with Logging {

  private val _minSplitsPerRDD = if (sparkSession.sparkContext.isLocal) {
    0 // will be split based on block by default.
  } else {
    math.max(readerOptions.hadoopConf.getInt("mapreduce.job.maps", 1),
      sparkSession.sparkContext.defaultMinPartitions)
  }

  SparkHadoopUtil.get.appendS3AndSparkHadoopConfigurations(
    sparkSession.sparkContext.getConf, readerOptions.hadoopConf)

  private val _broadcastedHadoopConf =
    sparkSession.sparkContext.broadcast(new SerializableConfiguration(readerOptions.hadoopConf))

  override def conf: SQLConf = sparkSession.sessionState.conf

  private def setReaderOptions(hiveAcidMetadata: HiveAcidMetadata) : Unit =  {
    // Push Down Predicate
    if (readerOptions.readConf.predicatePushdownEnabled) {
      setPushDownFiltersInHadoopConf(readerOptions.hadoopConf,
        hiveAcidMetadata,
        readerOptions.dataFilters)
    }

    // Set Required column.
    setRequiredColumnsInHadoopConf(readerOptions.hadoopConf,
      hiveAcidMetadata,
      readerOptions.requiredNonPartitionedColumns)

    logDebug(s"sarg.pushdown: " +
      s"${readerOptions.hadoopConf.get("sarg.pushdown")}," +
      s"hive.io.file.readcolumn.names: " +
      s"${readerOptions.hadoopConf.get("hive.io.file.readcolumn.names")}, " +
      s"hive.io.file.readcolumn.ids: " +
      s"${readerOptions.hadoopConf.get("hive.io.file.readcolumn.ids")}")
  }

  /**
   * @param hiveAcidMetadata - hive acid metadata for underlying table
   * @return - Returns RDD on top of non partitioned hive acid table and list of partitionNames empty list
    *         for entire table
   */
  def makeRDDForTable(hiveAcidMetadata: HiveAcidMetadata): RDD[InternalRow] = {
    setReaderOptions(hiveAcidMetadata)
    makeRDDForTable(
      hiveAcidMetadata.hTable,
      Util.classForName(hiveAcidOptions.tableDesc.getSerdeClassName,
        loadShaded = true).asInstanceOf[Class[Deserializer]],
      hiveAcidMetadata,
      readerOptions
    )
  }

  /**
   * @param hiveAcidMetadata - hive acid metadata of underlying table
   *
   * @return - Returns RDD on top of partitioned hive acid table
   */
  def makeRDDForPartitionedTable(hiveAcidMetadata: HiveAcidMetadata): RDD[InternalRow] = {

    val partitionToDeserializer = partitions.map(p => p.ptn.asInstanceOf[HiveJarPartition]).map {
      part =>
        val deserializerClassName = part.getTPartition.getSd.getSerdeInfo.getSerializationLib
        val deserializer = Util.classForName(deserializerClassName, loadShaded = true)
          .asInstanceOf[Class[Deserializer]]
        (part, deserializer)
    }.toMap
    makeRDDForPartitionedTable(partitionToDeserializer, filterOpt = None, readerOptions)
  }

  private def getSplitInfo (broadcastedConf: Broadcast[SerializableConfiguration],
                inputPathStr : String,
                validTxnList : String,
                validWriteId : String,
                ifcString : String,
                isFullAcidTable : Boolean,
                initLocalJobConfFuncOpt: Option[JobConf => Unit],
                requiredFields: Array[StructField],
                dataSchema: StructType,
                partitionSchema : StructType,
                mutableRow : InternalRow
               ) : HiveSplitInfoV2 = {
    HiveSplitInfoV2(0, broadcastedConf,
      inputPathStr, validTxnList, validWriteId,
      ifcString, isFullAcidTable, true,
      initLocalJobConfFuncOpt, requiredFields, dataSchema, partitionSchema, mutableRow, 0)
  }

  // TODO: This is performance hot spot we need to save as much time as possible.
  // Saving even a sec here can potentially save several mins in large dataset
  private def makeHiveSplitInfoForPartition(broadcastedConf: Broadcast[SerializableConfiguration],
                                    hiveAcidMetadata: HiveAcidMetadata,
                                    partition : HiveJarPartition,
                                    isFullAcidTable: Boolean,
                                    shouldCloneJobConf: Boolean,
                                    localTableDesc : TableDesc,
                                    tableParams: util.Map[String, String],
                                    requiredFields: Array[StructField],
                                    partitionSchema : StructType,
                                    colNames: String,
                                    colTypes: String,
                                    ifcString: String,
                                    inputPathStr: String,
                                    partitionKeyAttrs: Seq[(Attribute, Int)]) : HiveSplitInfoV2 = {
    val mutableRow = new SpecificInternalRow(hiveAcidMetadata.partitionSchema)
    val partSpec = partition.getSpec
    val partCols = partition.getTable.getPartitionKeys.asScala.map(_.getName)

    val partValues = if (partSpec == null) {
      Array.fill(partCols.size)(new String)
    } else {
      partCols.map(col => new String(partSpec.get(col))).toArray
    }

    //TODO: move this method to class scope
    def fillPartitionKeys(rawPartValues: Array[String], row: InternalRow): Unit = {
      var offset = 0
      partitionKeyAttrs.foreach { case (attr, ordinal) =>
        val partOrdinal = readerOptions.partitionAttributes.indexOf(attr)
        row(offset) = cast(
          Literal(rawPartValues(partOrdinal)), attr.dataType).eval(null)
        offset = offset + 1
      }
    }
    fillPartitionKeys(partValues, mutableRow)

    val initializeJobConfFunc = HiveAcidReader.initializeLocalJobConfFunc(
    inputPathStr, localTableDesc,
    tableParams,
    colNames, colTypes) _

    getSplitInfo(broadcastedConf,inputPathStr, validTxnList.writeToString(),
      validWriteIds.writeToString(), ifcString, isFullAcidTable, Some(initializeJobConfFunc),
      requiredFields, hiveAcidMetadata.dataSchema, partitionSchema, mutableRow)
  }

  // Scala tries to serialize the containing class along with method if method is
  // is being set to serialized object for executor.
  def makeV2InputPartition (split: HiveAcidPartition,
                            conf: Broadcast[SerializableConfiguration],
                            initLocalJobConfFuncOpt: Option[JobConf => Unit],
                            splitPath : String,
                            partitionValues : InternalRow,
                            requiredFields: Array[StructField],
                            dataSchema: StructType,
                            partitionSchema : StructType,
                            isFullAcidTable: Boolean): HiveAcidInputPartitionV2 = {
    new HiveAcidInputPartitionV2(split, conf, initLocalJobConfFuncOpt, splitPath,
      partitionValues, requiredFields, dataSchema, partitionSchema, isFullAcidTable)
  }

  def makeV2ReaderForTable(hiveAcidMetadata: HiveAcidMetadata): java.util.List[InputPartition[ColumnarBatch]] = {
    setReaderOptions(hiveAcidMetadata)
    val fieldSchemas = hiveAcidMetadata.hTable.getSd.getCols
    val colNames = getColumnNamesFromFieldSchema(fieldSchemas)
    val colTypes = getColumnTypesFromFieldSchema(fieldSchemas)

    val inputPathStr = hiveAcidMetadata.hTable.getPath.toString
    val ifc = hiveAcidMetadata.hTable.getInputFormatClass
      .asInstanceOf[java.lang.Class[InputFormat[Writable, Writable]]]
    val mutableRow = new SpecificInternalRow(hiveAcidMetadata.partitionSchema)
    val tableParams = hiveAcidMetadata.hTable.getParameters
    val tableDesc = hiveAcidMetadata.tableDesc

    val initializeJobConfFunc = HiveAcidReader.initializeLocalJobConfFunc(
      inputPathStr, tableDesc, tableParams, colNames, colTypes) _

    val requiredColumns = hiveAcidMetadata.tableSchema.fields.filter(field =>
      readerOptions.requiredNonPartitionedColumns.contains(field.name))

    val jobConf = HiveAcidCommon.getJobConf(readerOptions.hadoopConf,
      inputPathStr, true, Some(initializeJobConfFunc))

    val partitionArray = new java.util.ArrayList[InputPartition[ColumnarBatch]]

    val inputSplits  = HiveAcidCommon.getInputSplits(jobConf, validWriteIds,
      validTxnList.writeToString, ifc, hiveAcidMetadata.isFullAcidTable)

    for (i <- 0 until inputSplits.size)
      partitionArray.add(makeV2InputPartition(inputSplits(i),
        sparkSession.sparkContext.broadcast(new SerializableConfiguration(jobConf)),
        Some(initializeJobConfFunc), inputPathStr,
        mutableRow, requiredColumns, hiveAcidMetadata.dataSchema, hiveAcidMetadata.partitionSchema, hiveAcidMetadata.isFullAcidTable))

    partitionArray
  }

  private def makeHivePartitionInfoForV2(hiveAcidMetadata: HiveAcidMetadata): Seq[HiveSplitInfoV2] = {
    val partitionToDeserializer = getPartitionToDeserializer(partitions)
    setReaderOptions(hiveAcidMetadata)
    // Create and reuse values which are common for all partitions
    val requiredColumns = hiveAcidMetadata.tableSchema.fields.filter(field =>
      readerOptions.requiredNonPartitionedColumns.contains(field.name))
    val localTableDesc = hiveAcidOptions.tableDesc
    val partitionSchema = hiveAcidMetadata.partitionSchema;
    val tableParams = hiveAcidMetadata.hTable.getParameters

    val fieldSchemas = hiveAcidMetadata.hTable.getSd.getCols
    val colNames = getColumnNamesFromFieldSchema(fieldSchemas)
    val colTypes = getColumnTypesFromFieldSchema(fieldSchemas)
    val ifcName = hiveAcidMetadata.hTable.getInputFormatClass.getName

    val (partitionKeyAttrs, nonPartitionKeyAttrs) =
      readerOptions.requiredAttributes.zipWithIndex.partition { case (attr, _) =>
        readerOptions.partitionAttributes.contains(attr)
      }

    val broadcastCfg = sparkSession.sparkContext.broadcast(new SerializableConfiguration(readerOptions.hadoopConf))
    partitionToDeserializer.map { case (partition, partDeserializer) =>
      makeHiveSplitInfoForPartition(broadcastCfg, hiveAcidMetadata, partition,
        true, true, localTableDesc, tableParams, requiredColumns,
        partitionSchema, colNames, colTypes, ifcName,
        partition.getDataLocation.toString, partitionKeyAttrs)
    }.toSeq
  }

  def makeReaderForPartitionedTable(hiveAcidMetadata: HiveAcidMetadata):
    java.util.ArrayList[InputPartition[ColumnarBatch]] = {
    val partitionArray = new java.util.ArrayList[InputPartition[ColumnarBatch]]
    val splitsInfo = makeHivePartitionInfoForV2(hiveAcidMetadata)

    val partitionGenerator = new HiveAcidInputPartitionGeneratorV2(true, false)
    val parallelPartitionThreshold = sparkSession.sparkContext.getConf.getInt(SparkAcidConf.PARALLEL_PARTITION_THRESHOLD.configName, 20)
    val maxNoOfExecutors = sparkSession.sparkContext.getConf.getInt(SparkAcidConf.MAX_PARALLEL_PARTITION_FOR_SPLITS.configName, Int.MaxValue)

    val v2InputPartitions = if (splitsInfo.length > parallelPartitionThreshold) {
      val hiveSplitRDD = if (sparkSession.sparkContext.getConf.getBoolean(SparkAcidConf.ENABLE_CUSTOM_PARTITIONS_FOR_SPLITS.configName, true)) {
        val noOfExecutor = Math.min(maxNoOfExecutors, Math.ceil(splitsInfo.length.toDouble / parallelPartitionThreshold.toDouble).toInt)
        sparkSession.sparkContext.parallelize(splitsInfo, noOfExecutor)
      }
      else
        sparkSession.sparkContext.parallelize(splitsInfo)
      val partitonsReaders = partitionGenerator.generate(hiveSplitRDD)
      val cfg : Broadcast[SerializableConfiguration] = sparkSession.sparkContext.broadcast(new SerializableConfiguration(readerOptions.hadoopConf))
      partitonsReaders.asScala.foreach (ip => ip.setBroadcastConfig (cfg))
      partitonsReaders
    }
    else
      partitionGenerator.generateLocal(splitsInfo)

    partitionArray.addAll (v2InputPartitions)
    partitionArray
  }

  private def getPartitionToDeserializer(partitions: Seq[ReaderPartition])
                        : Map[HiveJarPartition, Class[_ <: Deserializer]] = {
    partitions.map(p => p.ptn.asInstanceOf[HiveJarPartition]).map {
      part =>
        val deserializerClassName = part.getTPartition.getSd.getSerdeInfo.getSerializationLib
        val deserializer = Util.classForName(deserializerClassName, loadShaded = true)
          .asInstanceOf[Class[Deserializer]]
        (part, deserializer)
    }.toMap
  }

  /**
   * Creates a Hadoop RDD to read data from the target table's data directory.
   * Returns a transformed RDD that contains deserialized rows.
   *
   * @param hiveTable Hive metadata for the table being scanned.
   * @param deserializerClass Class of the SerDe used to deserialize Writables read from Hadoop.
   */
  private def makeRDDForTable(hiveTable: HiveTable,
                              deserializerClass: Class[_ <: Deserializer],
                              hiveAcidMetadata: HiveAcidMetadata,
                              readerOptions: ReaderOptions): RDD[InternalRow] = {

    assert(!hiveTable.isPartitioned,
      "makeRDDForTable() cannot be called on a partitioned table, since input formats may " +
        "differ across partitions. Use makeRDDForPartitionedTable() instead.")

    // Create local references to member variables, so that the entire `this` object won't be
    // serialized in the closure below.
    val localTableDesc = hiveAcidOptions.tableDesc
    val tablePath = hiveTable.getPath

    val ifcName = hiveTable.getInputFormatClass.getName
    val ifc = Util.classForName(ifcName, loadShaded = true)
      .asInstanceOf[java.lang.Class[InputFormat[Writable, Writable]]]
    val hiveRDD = createRddForTable(localTableDesc, hiveTable.getSd.getCols,
      hiveTable.getParameters, tablePath.toString, ifc)
    deserializeTableRdd(hiveRDD, deserializerClass)
  }

  /**
    * Utility to convert hiveRDD on non-partitioned table to Deserialized RDD which transforms
    * raw Records from hiveRDD to InternalRow
    * @param hiveRDD
    * @param deserializerClass
    * @return Deserialized RDD
    */
  private def deserializeTableRdd(hiveRDD: RDD[(RecordIdentifier, Writable)],
                                  deserializerClass: Class[_ <: Deserializer]) = {
    val localTableDesc = hiveAcidOptions.tableDesc
    val broadcastedHadoopConf = _broadcastedHadoopConf
    val attrsWithIndex = readerOptions.requiredAttributes.zipWithIndex
    val outputRowDataTypes = readerOptions.requiredAttributes.map(_.dataType)
    val mutableRow = new SpecificInternalRow(outputRowDataTypes)

    val shouldIncludeRowIds = readerOptions.readConf.includeRowIds &&
      readerOptions.requiredAttributes.exists(HiveAcidReader.isAttributeRowId(_))
    val mutableRowRecordIds = if (shouldIncludeRowIds) {
      Some(new SpecificInternalRow(HiveAcidMetadata.rowIdSchema.fields.map(_.dataType)))
    } else {
      None
    }

    // ReaderOptions cannot be serialized
    val deserializedHiveRDD = hiveRDD.mapPartitions { iter =>
      val hconf = broadcastedHadoopConf.value.value
      val deserializer = deserializerClass.newInstance()
      deserializer.initialize(hconf, localTableDesc.getProperties)
      HiveAcidReader.fillObject(iter, deserializer, attrsWithIndex, mutableRow,
        mutableRowRecordIds,
        deserializer)
    }
    new HiveAcidUnionRDD[InternalRow](sparkSession.sparkContext, Seq(deserializedHiveRDD), Seq())
  }

  /**
   * Create a HiveAcidRDD for every partition key specified in the query.
   *
   * @param partitionToDeserializer Mapping from a Hive Partition metadata object to the SerDe
   *     class to use to deserialize input Writables from the corresponding partition.
   * @param filterOpt If defined, then the filter is used to reject files contained in the data
   *     subdirectory of each partition being read. If None, then all files are accepted.
   */
  private def makeRDDForPartitionedTable(
      partitionToDeserializer: Map[HiveJarPartition, Class[_ <: Deserializer]],
      filterOpt: Option[PathFilter],
      readerOptions: ReaderOptions): RDD[InternalRow] = {

    val hivePartitionRDDSeq = partitionToDeserializer.map { case (partition, partDeserializer) =>
      val partPath = partition.getDataLocation
      val inputPathStr = applyFilterIfNeeded(partPath, filterOpt)
      val ifcString = partition.getTPartition.getSd.getInputFormat
      val ifc = Util.classForName(ifcString, loadShaded = true)
        .asInstanceOf[java.lang.Class[InputFormat[Writable, Writable]]]

      // Create local references so that the outer object isn't serialized.
      val localTableDesc = hiveAcidOptions.tableDesc
      val rdd = createRddForTable(localTableDesc, partition.getCols,
        partition.getTable.getParameters, inputPathStr, ifc)
      val hiveSplitInfo = rdd.asInstanceOf[HiveAcidRDD[Writable, Writable]].getHiveSplitsInfo
      (rdd, partition, partDeserializer, hiveSplitInfo)
    }.toSeq

    val hivePartitionRDDs = hivePartitionRDDSeq.map {
      case (hiveRDD, partition, partDeserializer, _)  => {
        makeRddForPartition(hiveRDD, partition, partDeserializer, readerOptions)
      }
    }

    // Even if we don't use any partitions, we still need an empty RDD
    if (hivePartitionRDDs.isEmpty) {
      new EmptyRDD[InternalRow](sparkSession.sparkContext)
    } else {
      val hiveSplitInfos = hivePartitionRDDSeq.map(_._4)
      new HiveAcidUnionRDD[InternalRow](hivePartitionRDDs.head.context, hivePartitionRDDs, hiveSplitInfos)
    }
  }

  private def makeRddForPartition(hiveRDD: RDD[(RecordIdentifier, Writable)],
                                  partition: HiveJarPartition,
                                  partDeserializer: Class[_ <: Deserializer],
                                  readerOptions: ReaderOptions) = {
    deserializePartitionRdd(hiveRDD, partition, partDeserializer)
  }

  /**
    * Convert Hive RDD for a particular partition into Deserialized RDD which transforms every Raw row
    * into InternalRow.
    *
    * @param partitionRDD
    * @param partition
    * @param partDeserializer
    * @return
    */
  private def deserializePartitionRdd(partitionRDD: RDD[(RecordIdentifier, Writable)], partition: HiveJarPartition, partDeserializer: Class[_ <: Deserializer]) = {
    // member variable cannot be used directly inside mapPartition as HiveAcidReader is not serializable.
    val broadcastedHadoopConf = _broadcastedHadoopConf
    val tableProperties = hiveAcidOptions.tableDesc.getProperties
    val partProps = partition.getMetadataFromPartitionSchema
    val localTableDesc = hiveAcidOptions.tableDesc
    val outputRowDataTypes = readerOptions.requiredAttributes.map(_.dataType)
    // Get partition field info
    val partSpec = partition.getSpec
    val partCols = partition.getTable.getPartitionKeys.asScala.map(_.getName)
    // 'partValues[i]' contains the value for the partitioning column at 'partCols[i]'.
    val partValues = if (partSpec == null) {
      Array.fill(partCols.size)(new String)
    } else {
      partCols.map(col => new String(partSpec.get(col))).toArray
    }
    val mutableRow = new SpecificInternalRow(outputRowDataTypes)
    val shouldIncludeRowIds = readerOptions.readConf.includeRowIds &&
      readerOptions.requiredAttributes.exists(HiveAcidReader.isAttributeRowId(_))
    val mutableRowRecordIds = if (shouldIncludeRowIds) {
      Some(new SpecificInternalRow(HiveAcidMetadata.rowIdSchema.fields.map(_.dataType)))
    } else {
      None
    }

    // Splits all attributes into two groups, partition key attributes and those
    // that are not. Attached indices indicate the position of each attribute in
    // the output schema.
    val (partitionKeyAttrs, nonPartitionKeyAttrs) =
    readerOptions.requiredAttributes.zipWithIndex.partition { case (attr, _) =>
      readerOptions.partitionAttributes.contains(attr)
    }

    def fillPartitionKeys(rawPartValues: Array[String], row: InternalRow, includeRowIds: Boolean): Unit = {
      partitionKeyAttrs.foreach { case (attr, ordinal) =>
        val partOrdinal = readerOptions.partitionAttributes.indexOf(attr)
        row(ordinal) = cast(
          Literal(rawPartValues(partOrdinal)), attr.dataType).eval(null)
      }
    }

    // Fill all partition keys to the given MutableRow object
    fillPartitionKeys(partValues, mutableRow, shouldIncludeRowIds)


    partitionRDD.mapPartitions { iter =>
      val hconf = broadcastedHadoopConf.value.value
      val deserializer = partDeserializer.newInstance()
      // SPARK-13709: For SerDes like AvroSerDe, some essential information (e.g. Avro schema
      // information) may be defined in table properties. Here we should merge table properties
      // and partition properties before initializing the deserializer. Note that partition
      // properties take a higher priority here. For example, a partition may have a different
      // SerDe as the one defined in table properties.
      val props = new Properties(tableProperties)
      partProps.asScala.foreach {
        case (key, value) => props.setProperty(key, value)
      }
      deserializer.initialize(hconf, props)

      // get the table deserializer
      val tableSerDeClassName = localTableDesc.getSerdeClassName
      val tableSerDe = Util.classForName(tableSerDeClassName,
        loadShaded = true).newInstance().asInstanceOf[Deserializer]
      tableSerDe.initialize(hconf, localTableDesc.getProperties)

      // fill the non partition key attributes
      HiveAcidReader.fillObject(iter, deserializer, nonPartitionKeyAttrs,
        mutableRow,
        mutableRowRecordIds,
        tableSerDe)
    }
  }

  /**
    * If `filterOpt` is defined, then it will be used to filter files from `path`. These files are
    * returned in a single, comma-separated string.
    */
  private def applyFilterIfNeeded(path: Path, filterOpt: Option[PathFilter]): String = {
    filterOpt match {
      case Some(filter) =>
        val fs = path.getFileSystem(readerOptions.hadoopConf)
        val filteredFiles = fs.listStatus(path, filter).map(_.getPath.toString)
        filteredFiles.mkString(",")
      case None => path.toString
    }
  }

  /**
    * Creates a HiveAcidRDD based on the broadcasted HiveConf and other job properties that will be
    * applied locally on each slave.
    */
  private def createRddForTable(tableDesc: TableDesc,
                                cols: util.List[FieldSchema],
                                tableParameters: util.Map[String, String],
                                path: String,
                                inputFormatClass: Class[InputFormat[Writable, Writable]]
                               ): RDD[(RecordIdentifier, Writable)] = {

    val colNames = getColumnNamesFromFieldSchema(cols)
    val colTypes = getColumnTypesFromFieldSchema(cols)
    val initializeJobConfFunc = HiveAcidReader.initializeLocalJobConfFunc(
      path, tableDesc, tableParameters,
      colNames, colTypes) _

    val init = HiveAcidReader.initializeLocalJobConfFunc(
      path, tableDesc,
      tableParameters,
      colNames, colTypes) _

    val rdd = new HiveAcidRDD(
      sparkSession,
      validWriteIds,
      validTxnList,
      hiveAcidOptions.isFullAcidTable,
      _broadcastedHadoopConf.asInstanceOf[Broadcast[SerializableConfiguration]],
      Some(initializeJobConfFunc),
      inputFormatClass,
      classOf[Writable],
      classOf[Writable],
      _minSplitsPerRDD,
      tableDesc)

    rdd
  }

  private def setRequiredColumnsInHadoopConf(conf: Configuration,
                                     acidTableMetadata: HiveAcidMetadata,
                                     requiredColumns: Seq[String]): Unit = {
    val dataCols: Seq[String] = acidTableMetadata.dataSchema.fields.map(_.name)
    val requiredColumnIndexes = requiredColumns.map(a => dataCols.indexOf(a): Integer)
    val (sortedIDs, sortedNames) = requiredColumnIndexes.zip(requiredColumns).sorted.unzip
    conf.set(ColumnProjectionUtils.READ_ALL_COLUMNS, "false")
    conf.set(ColumnProjectionUtils.READ_COLUMN_NAMES_CONF_STR, sortedNames.mkString(","))
    conf.set(ColumnProjectionUtils.READ_COLUMN_IDS_CONF_STR, sortedIDs.mkString(","))
  }

  private def setPushDownFiltersInHadoopConf(conf: Configuration,
                                     acidTableMetadata: HiveAcidMetadata,
                                     dataFilters: Array[Filter]): Unit = {
    HiveAcidSearchArgument.build(acidTableMetadata.dataSchema, dataFilters).foreach { f =>
      logDebug(s"searchArgument: $f")
      conf.set("sarg.pushdown", ConvertAstToSearchArg.sargToKryo(f))
      conf.setBoolean(ConfVars.HIVEOPTINDEXFILTER.varname, true)
    }
  }
}

private[reader] object HiveAcidReader extends Hive3Inspectors with Logging {

  def getPartitions(hiveAcidMetadata: HiveAcidMetadata,
                    readerOptions: ReaderOptions,
                    partitionFilters: Seq[Filter]): (Seq[ReaderPartition], Seq[String]) = {

    val partitions = if (hiveAcidMetadata.isPartitioned) {
      if (readerOptions.readConf.metastorePartitionPruningEnabled) {
        val partitionPruningFiltering = Option(HiveConverter.compileFilters(partitionFilters))
        try {
          hiveAcidMetadata.getRawPartitions(partitionPruningFiltering)
        } catch {
          // TODO: Enable pruning results returned by getting all Partitions
          case ex: org.apache.hadoop.hive.metastore.api.MetaException => {
            logWarning("Caught Hive MetaException attempting to get partition metadata by " +
              "filter from Hive. Falling back to fetching all partition metadata and pruning them. " +
              "Filter: " + partitionPruningFiltering)
            logDebug("Exception: " + ex.getMessage, ex)
            val allPartitions = hiveAcidMetadata.getRawPartitions()
            try {
              prunePartitions(allPartitions, hiveAcidMetadata, readerOptions, partitionPruningFiltering)
            } catch {
              case e: Exception =>
                logWarning("Error while pruning the partitions. All the partitions will be returned" +
                  " that can cause Performance impact", e)
                allPartitions
            }
          }
        }
      } else {
        hiveAcidMetadata.getRawPartitions()
      }
    } else {
      Seq()
    }

    val partitionNames = partitions.map(_.getName())

    (partitions.map(p => ReaderPartition(p)), partitionNames)
  }

  /**
    * This will try to fetch all Partitions and prune them according to filter specified
    * @return prunedPartitions
    */
  private def prunePartitions(partitions: Seq[HiveJarPartition],
                              hiveAcidMetadata: HiveAcidMetadata,
                              readerOptions: ReaderOptions,
                              partitionPruningFiltering: Option[String]): Seq[HiveJarPartition] = {
    partitionPruningFiltering match {
      case Some(filter) => {
        val filterExp = functions.expr(filter).expr
        val partitionMap: Map[CatalogTablePartition, HiveJarPartition] =
          partitions.map(hp => (HiveAcidUtils.convertToCatalogTablePartition(hp), hp)).toMap
        val catalogParts = HiveAcidUtils.prunePartitionsByFilter(hiveAcidMetadata,
          partitionMap.keys.toSeq,
          Some(filterExp),
          readerOptions.sessionLocalTimeZone)
        val partCols = hiveAcidMetadata.partitionSchema.fieldNames
        catalogParts.map {
          cPart =>
            val partValues = partCols.map { hc =>
              cPart.spec.get(hc).getOrElse {
                throw new IllegalArgumentException(
                  s"Partition spec is missing a value for column '${hc}': ${cPart.spec}")
              }
            }
            partitionMap.get(cPart).getOrElse {
              throw new IllegalArgumentException(
                s"Partition values is missing from raw partitions ${partValues}"
              )
            }
        }
      }
      case _ => partitions
    }
  }

  // copied from PlanUtils.configureJobPropertiesForStorageHandler(tableDesc)
  // that calls Hive.get() which tries to access metastore, but it's not valid in runtime
  // it would be fixed in next version of hive but till then, we should use this instead
  private def configureJobPropertiesForStorageHandler(tableDesc: TableDesc,
                                                      conf: Configuration, input: Boolean) {
    val property = tableDesc.getProperties.getProperty(META_TABLE_STORAGE)
    val storageHandler =
      org.apache.hadoop.hive.ql.metadata.HiveUtils.getStorageHandler(conf, property)
    if (storageHandler != null) {
      val jobProperties = new java.util.LinkedHashMap[String, String]
      if (input) {
        storageHandler.configureInputJobProperties(tableDesc, jobProperties)
      } else {
        storageHandler.configureOutputJobProperties(tableDesc, jobProperties)
      }
      if (!jobProperties.isEmpty) {
        tableDesc.setJobProperties(jobProperties)
      }
    }
  }

  /**
    * Curried. After given an argument for 'path', the resulting JobConf => Unit closure is used to
    * instantiate a HiveAcidRDD.
    */
  def initializeLocalJobConfFunc(path: String, tableDesc: TableDesc,
                                 tableParameters: util.Map[String, String],
                                 schemaColNames: String,
                                 schemaColTypes: String)(jobConf: JobConf) {
    FileInputFormat.setInputPaths(jobConf, Seq[Path](new Path(path)): _*)
    if (tableDesc != null) {
      configureJobPropertiesForStorageHandler(tableDesc, jobConf, input = true)
      Utilities.copyTableJobPropertiesToConf(tableDesc, jobConf)
    }
    val bufferSize = System.getProperty("spark.buffer.size", "65536")
    jobConf.set("io.file.buffer.size", bufferSize)
    jobConf.set("schema.evolution.columns", schemaColNames)
    jobConf.set("schema.evolution.columns.types", schemaColTypes)

    // Set HiveACID related properties in jobConf
    val isTransactionalTable = AcidUtils.isTransactionalTable(tableParameters)
    val acidProps = if (isTransactionalTable) {
      AcidUtils.getAcidOperationalProperties(tableParameters)
    } else {
      null
    }
    AcidUtils.setAcidOperationalProperties(jobConf, isTransactionalTable, acidProps)
  }

  /**
    * Transform all given raw `(RowIdentifier, Writable)`s into `InternalRow`s.
    *
    * @param iterator Iterator of all `Writable`s to be transformed
    * @param rawDeser The `Deserializer` associated with the input `Writable`
    * @param nonPartitionKeyAttrs Attributes that should be filled together with their corresponding
    *                             positions in the output schema
    * @param mutableRow A reusable `MutableRow` that should be filled
    * @param tableDeser Table Deserializer
    * @return An `Iterator[Row]` transformed from `iterator`
    */
  def fillObject(
                  iterator: Iterator[(RecordIdentifier, Writable)],
                  rawDeser: Deserializer,
                  nonPartitionKeyAttrs: Seq[(Attribute, Int)],
                  mutableRow: InternalRow,
                  mutableRowRecordId: Option[InternalRow],
                  tableDeser: Deserializer): Iterator[InternalRow] = {

    // Note mutableRowRecordId will be None when no rowIds needs to be included in the InternalRow.
    val soi = if (rawDeser.getObjectInspector.equals(tableDeser.getObjectInspector)) {
      rawDeser.getObjectInspector.asInstanceOf[StructObjectInspector]
    } else {
      ObjectInspectorConverters.getConvertedOI(
        rawDeser.getObjectInspector,
        tableDeser.getObjectInspector).asInstanceOf[StructObjectInspector]
    }

    logDebug(soi.toString)

    val (fieldRefs, fieldOrdinals) = if (!mutableRowRecordId.isEmpty) { // Check if we need to include rowIds
      nonPartitionKeyAttrs.filterNot(x => isAttributeRowId(x._1)
        && soi.getStructFieldRef(x._1.name) == null).map {
        case (attr, ordinal) => soi.getStructFieldRef(attr.name) -> ordinal
      }.toArray.unzip
    } else {
      nonPartitionKeyAttrs.map { case (attr, ordinal) =>
        soi.getStructFieldRef(attr.name) -> ordinal
      }.toArray.unzip
    }

    // Builds specific unwrappers ahead of time according to object inspector
    // types to avoid pattern matching and branching costs per row.
    val unwrappers: Seq[(Any, InternalRow, Int) => Unit] = fieldRefs.map {
      x =>
        val y = x.getFieldObjectInspector
        y match {
          case oi: BooleanObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setBoolean(ordinal, oi.get(value))
          case oi: ByteObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setByte(ordinal, oi.get(value))
          case oi: ShortObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setShort(ordinal, oi.get(value))
          case oi: IntObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setInt(ordinal, oi.get(value))
          case oi: LongObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setLong(ordinal, oi.get(value))
          case oi: FloatObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setFloat(ordinal, oi.get(value))
          case oi: DoubleObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) => row.setDouble(ordinal, oi.get(value))
          case oi: HiveVarcharObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              row.update(ordinal, UTF8String.fromString(oi.getPrimitiveJavaObject(value).getValue))
          case oi: HiveCharObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              row.update(ordinal, UTF8String.fromString(oi.getPrimitiveJavaObject(value).getValue))
          case oi: HiveDecimalObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              row.update(ordinal, toCatalystDecimal(oi, value))
          case oi: TimestampObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              // convert hive timestamp from UTC to JVM timezone to make it consistent with spark
              // this makes direct reader V1 consistent with spark-native for ORC, AVRO and Text formats.
              val converted  = HWCDateTimeUtils.hiveTimestampToMicrosForDisplay(oi.getPrimitiveJavaObject(value))
              row.setLong(ordinal, converted)
          case oi: DateObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              val y = oi.getPrimitiveWritableObject(value).get().toEpochMilli
              row.setInt(ordinal, DateTimeUtils.fromJavaDate(new java.sql.Date(y)))
          case oi: BinaryObjectInspector =>
            (value: Any, row: InternalRow, ordinal: Int) =>
              row.update(ordinal, oi.getPrimitiveJavaObject(value))
          case oi =>
            val unwrapper = unwrapperFor(oi)
            (value: Any, row: InternalRow, ordinal: Int) => row(ordinal) = unwrapper(value)
        }
    }

    val converter = ObjectInspectorConverters.getConverter(rawDeser.getObjectInspector, soi)

    // Map each tuple to a row object
    // Find rowIdIndex
    val rowIdIndex = nonPartitionKeyAttrs.find(attrIdx => isAttributeRowId(attrIdx._1))
      .getOrElse((None, -1))._2
    if (rowIdIndex < 0 && !mutableRowRecordId.isEmpty) {
      HiveAcidErrors.unexpectedReadError("Error reading the ACID table" +
        " as rowId is not present even when includeRowId is the parameter")
    }
    iterator.map { value =>
      val raw = converter.convert(rawDeser.deserialize(value._2))
      var i = 0
      mutableRowRecordId match {
        case Some(record) =>
          val recordIdentifier = value._1
          record.setLong(0, recordIdentifier.getWriteId)
          record.setInt(1, recordIdentifier.getBucketProperty)
          record.setLong(2, recordIdentifier.getRowId)
          mutableRow.update(rowIdIndex, record)
        case None =>
      }
      val length = fieldRefs.length
      while (i < length) {
        val fieldValue = soi.getStructFieldData(raw, fieldRefs(i))
        if (fieldValue == null) {
          mutableRow.setNullAt(fieldOrdinals(i))
        } else {
          unwrappers(i)(fieldValue, mutableRow, fieldOrdinals(i))
        }
        i += 1
      }
      mutableRow: InternalRow
    }
  }

  def isAttributeRowId(attribute: Attribute): Boolean = {
    attribute.dataType.typeName == "struct" && attribute.name == "rowId"
  }
}
