/*
 * Copyright 2020 Cloudera, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.qubole.spark.hiveacid.util

import com.qubole.spark.hiveacid.reader.v2.HiveAcidInputPartitionV2
import org.apache.hadoop.hive.common.ValidReaderWriteIdList
import org.apache.hadoop.io.Writable
import org.apache.hadoop.mapred.{InputFormat, InvalidInputException, JobConf}
import org.apache.spark.broadcast.Broadcast
import org.apache.spark.deploy.SparkHadoopUtil
import org.apache.spark.internal.Logging
import org.apache.spark.rdd.RDD
import org.apache.spark.sql.catalyst.InternalRow
import org.apache.spark.sql.types.{StructField, StructType}

case class HiveSplitInfoV2(id : Int,
                           broadcastedConf: Broadcast[SerializableConfiguration],
                           splitPath : String,
                           validTxnList : String,
                           validWriteIdList: String,
                           ifcName: String,
                           isFullAcidTable: Boolean,
                           shouldCloneJobConf: Boolean,
                           initLocalJobConfFuncOpt: Option[JobConf => Unit],
                           requiredFields: Array[StructField],
                           dataSchema: StructType,
                           partitionSchema : StructType,
                           partValues : InternalRow,
                           minPartitions: Int)

private[hiveacid] case class HiveAcidInputPartitionGeneratorV2 (ignoreEmptySplits: Boolean,
                                                                ignoreMissingFiles: Boolean) extends Logging {
  def getPartitions[K, V](jobConf: JobConf,
                          validTxnList : String,
                          validWriteIdList : String,
                          inputFormat: Class[InputFormat[Writable, Writable]],
                          minPartitions: Int,
                          isFullAcidTable : Boolean,
                          partValues: InternalRow,
                          requiredFields: Array[StructField],
                          dataSchema: StructType,
                          partitionSchema : StructType,
                          broadcastedConf: Broadcast[SerializableConfiguration],
                          initLocalJobConfFuncOpt: Option[JobConf => Unit],
                          splitPath : String) : Array[HiveAcidInputPartitionV2] = {
    // add the credentials here as this can be called before SparkContext initialized
    SparkHadoopUtil.get.addCredentials(jobConf)
    try {
      val allInputSplits = HiveAcidCommon.getInputSplits (jobConf, new ValidReaderWriteIdList(validWriteIdList), validTxnList, inputFormat, isFullAcidTable, minPartitions, true, true)
      val array = new Array[HiveAcidInputPartitionV2](allInputSplits.length)
      for (i <- allInputSplits.indices) {
        array (i) = new HiveAcidInputPartitionV2(allInputSplits(i),
          broadcastedConf, initLocalJobConfFuncOpt, splitPath,
          partValues, requiredFields, dataSchema, partitionSchema, isFullAcidTable)
      }
      array
    } catch {
      case e: InvalidInputException if ignoreMissingFiles =>
        val inputDir = jobConf.get(org.apache.hadoop.mapreduce.lib.input.FileInputFormat.INPUT_DIR)
        logWarning(s"$inputDir doesn't exist and no" +
          s" partitions returned from this path.", e)
        Array.empty[HiveAcidInputPartitionV2]
    }
  }

  def generateLocal(splitsInfoRDD: Seq[HiveSplitInfoV2]): java.util.ArrayList[HiveAcidInputPartitionV2] = {
    val splits = splitsInfoRDD.map {
      case HiveSplitInfoV2(id, broadcastedConf, splitPath, validTxnList,
      validWriteIdList, ifcName, isFullAcidTable, shouldCloneJobConf, initLocalJobConfFuncOpt, requiredFields, dataSchema, partitionSchema, partValues, minPartitions) =>
        val jobConf = HiveAcidCommon.getJobConf(broadcastedConf.value.value,
          splitPath, true, initLocalJobConfFuncOpt)
        getPartitions[Writable, Writable](jobConf, validTxnList, validWriteIdList, HiveAcidInputPartitionGeneratorV2.getInputFormat(ifcName),
          minPartitions, isFullAcidTable, partValues, requiredFields, dataSchema, partitionSchema, broadcastedConf, initLocalJobConfFuncOpt, splitPath)
    }
    val partitions = new java.util.ArrayList[HiveAcidInputPartitionV2]
    splits.foreach {
      _.foreach {
        partitions.add(_)
      }
    }
    partitions
  }

  def generate(splitsInfoRDD: RDD[HiveSplitInfoV2]): java.util.ArrayList[HiveAcidInputPartitionV2] = {
    logInfo("Spawning job to compute partitions for ACID table RDD")
    val splits = splitsInfoRDD.map {
      case HiveSplitInfoV2(id, broadcastedConf, splitPath, validTxnList,
      validWriteIdList, ifcName, isFullAcidTable, shouldCloneJobConf, initLocalJobConfFuncOpt, requiredFields, dataSchema, partitionSchema, partValues, minPartitions) =>
        val jobConf = HiveAcidCommon.getJobConf(broadcastedConf.value.value,
        splitPath, true, initLocalJobConfFuncOpt)
        getPartitions[Writable, Writable](jobConf, validTxnList, validWriteIdList, HiveAcidInputPartitionGeneratorV2.getInputFormat(ifcName),
          minPartitions, isFullAcidTable, partValues, requiredFields, dataSchema, partitionSchema, broadcastedConf,  initLocalJobConfFuncOpt, splitPath)
    }.collect()
    val partitions = new java.util.ArrayList[HiveAcidInputPartitionV2]
    splits.foreach {
      _.foreach {
        partitions.add(_)
      }
    }
    partitions
  }
}
private[hiveacid] object HiveAcidInputPartitionGeneratorV2 extends Logging {
  private def getInputFormat(inputFormatClassName: String): Class[InputFormat[Writable, Writable]] = {
    Util.classForName(inputFormatClassName, false)
      .asInstanceOf[java.lang.Class[InputFormat[Writable, Writable]]]
  }
}
