/*
 * Copyright 2019 Qubole, Inc.  All rights reserved.
 *
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.qubole.spark.hiveacid.writer

import com.qubole.spark.hiveacid.{HiveAcidErrors, HiveAcidOperation, SparkAcidConf}
import com.qubole.spark.hiveacid.hive.HiveAcidMetadata
import com.qubole.spark.hiveacid.transaction.HiveAcidTxn
import com.qubole.spark.hiveacid.util.SerializableConfiguration
import com.qubole.spark.hiveacid.writer.hive.{HiveAcidFullAcidWriter, HiveAcidInsertOnlyWriter, HiveAcidWriterOptions}
import org.apache.log4j.{Logger, LogManager}
import org.apache.spark.internal.Logging
import org.apache.spark.sql.{DataFrame, SparkSession}
import org.apache.spark.sql.catalyst.{InternalRow, TableIdentifier}
import org.apache.spark.sql.catalyst.catalog.CatalogTypes.TablePartitionSpec
import org.apache.spark.sql.catalyst.expressions.Attribute
import org.apache.spark.sql.execution.command.AlterTableAddPartitionCommand
import org.apache.spark.sql.execution.datasources.PartitioningUtils
import org.apache.spark.sql.functions.col
import org.apache.spark.sql.types.StructType
import scala.collection.JavaConverters.mapAsScalaMapConverter
import scala.language.implicitConversions

/**
 * Performs eager write of a dataframe df to a hive acid table based on operationType
 *
 * @param sparkSession    - Spark session
 * @param curTxn           - Transaction object to acquire locks.
 * @param hiveAcidMetadata - Hive acid table where we want to write dataframe
 * @param statementId      - Optional. In a same transaction, multiple statements like INSERT/UPDATE/DELETE
 *                           (like in case of MERGE) can be issued.
 *                           [[statementId]] have to be different for them to ensure delta collision
 *                           is avoided for them during writes.
 */
private[hiveacid] class TableWriter(sparkSession: SparkSession,
                                    curTxn: HiveAcidTxn,
                                    hiveAcidMetadata: HiveAcidMetadata,
                                    sparkAcidConf: SparkAcidConf,
                                    statementId: Option[Int] = None) extends Logging {

  private val MAX_NUMBER_OF_BUCKETS = 4096

  private def getSchema(operationType: HiveAcidOperation.OperationType): StructType = {
    val expectRowIdsInDataFrame = operationType match {
      case HiveAcidOperation.INSERT_OVERWRITE | HiveAcidOperation.INSERT_INTO => false
      case HiveAcidOperation.DELETE | HiveAcidOperation.UPDATE => true
      case _ => throw HiveAcidErrors.invalidOperationType(operationType.toString)
    }

    if (expectRowIdsInDataFrame) {
      hiveAcidMetadata.tableSchemaWithRowId
    } else {
      hiveAcidMetadata.tableSchema
    }
  }

  private def getColumns(operationType: HiveAcidOperation.OperationType,
            df: DataFrame): (Seq[Attribute], Array[Attribute], Seq[Attribute]) = {

    val columnNames = getSchema(operationType).fields.map(_.name)

    val allColumns = df.queryExecution.optimizedPlan.output.zip(columnNames).map {
      case (attr, columnName) =>
        attr.withName(columnName)
    }

    val allColumnNameToAttrMap = allColumns.map(attr => attr.name -> attr).toMap

    val partitionColumns = hiveAcidMetadata.partitionSchema.fields.map(
      field => allColumnNameToAttrMap(field.name))

    val dataColumns = allColumns.filterNot(partitionColumns.contains)

    (allColumns, partitionColumns, dataColumns)
  }

  /**
    * Common utility function to process all types of operations insert/update/delete
    * for the hive acid table
    * @param operationType type of operation.
    * @param df data frame to be written into the table.
    */
  def process(operationType: HiveAcidOperation.OperationType,
              df: DataFrame): Unit = {

    val hadoopConf = sparkSession.sessionState.newHadoopConf()

    val (allColumns, partitionColumns, dataColumns) = getColumns(operationType, df)

    var isWriteEnable: Boolean = false
    isWriteEnable = sparkSession.sessionState.conf.getConfString("tests.enable.write.via.direct.reader.mode", "false") == "true"
    val logger: Logger = LogManager.getLogger(this.getClass)

    if (!isWriteEnable) {
      logger.error("Write or Insert into table, through spark-acid is not supported please use HWC")
      throw new UnsupportedOperationException("Write or Insert into table, through spark-acid is not supported please use HWC")
    }

    try {

      logger.warn("Write through spark-acid is enabled for testing only and is not recommended for production. [isWriteEnabled= "+isWriteEnable+"]")
      // FIXME: IF we knew the partition then we should
      //   only lock that partition.
      curTxn.addTableLock(hiveAcidMetadata.dbName, hiveAcidMetadata.tableName)
      curTxn.acquireLocks(operationType, sparkAcidConf)

      if (!HiveAcidTxn.IsTxnStillValid(curTxn, hiveAcidMetadata.fullyQualifiedName)) {
        logInfo(s"Transaction ${curTxn.txnId} is no more valid for table ${hiveAcidMetadata.fullyQualifiedName} as other" +
          s" transaction might have got committed before lock got acquired")
        HiveAcidErrors.txnOutdated(curTxn.txnId, hiveAcidMetadata.fullyQualifiedName)
      }

      // Create Snapshot !!!
      val curSnapshot = HiveAcidTxn.createSnapshot(curTxn, hiveAcidMetadata)

      val writerOptions = new WriterOptions(curSnapshot.currentWriteId,
        operationType,
        new SerializableConfiguration(hadoopConf),
        getSchema(operationType),
        dataColumns,
        partitionColumns,
        allColumns,
        sparkSession.sessionState.conf.sessionLocalTimeZone,
        statementId
      )

      val isFullAcidTable = hiveAcidMetadata.isFullAcidTable

      val hiveAcidWriterOptions = HiveAcidWriterOptions.get(hiveAcidMetadata, writerOptions)

      // This RDD is serialized and sent for distributed execution.
      // All the access object in this needs to be serializable.
      val processRddPartition = new (Iterator[InternalRow] => Seq[TablePartitionSpec]) with
        Serializable {
        override def apply(iterator: Iterator[InternalRow]): Seq[TablePartitionSpec] = {
          val writer = if (isFullAcidTable) {
            new HiveAcidFullAcidWriter(writerOptions, hiveAcidWriterOptions)
          } else {
            new HiveAcidInsertOnlyWriter(writerOptions, hiveAcidWriterOptions)
          }
          iterator.foreach { row => writer.process(row) }
          writer.close()
          writer.partitionsTouched()
        }
      }

      val resultRDD =
        operationType match {
          // In order to read data from delete delts, hive uses mergesort, which requires
          // originalWriteId, bucket, and rowId in ascending order, and currentWriteId in descending order.
          // We take care of originalWriteId, bucket, and rowId in asc order here. We only write file per bucket-transaction,
          // hence currentWriteId remains same throughout the file and doesn't need ordering.
          //
          // Deleted rowId needs to be in same bucketed file name as the original row. To achieve this,
          // we repartition into 4096 partitions (i.e maximum number of buckets).
          // This ensures all rows of one bucket goes to same partition.
          //
          // There is still a chance that rows from multiple buckets go to same partition as well, but this is expected to work!
          case HiveAcidOperation.DELETE | HiveAcidOperation.UPDATE =>
            df.repartition(MAX_NUMBER_OF_BUCKETS, col("rowId.bucketId"))
              .toDF.sortWithinPartitions("rowId.bucketId", "rowId.writeId", "rowId.rowId")
              .toDF.queryExecution.executedPlan.execute()
          case HiveAcidOperation.INSERT_OVERWRITE | HiveAcidOperation.INSERT_INTO =>
            df.queryExecution.executedPlan.execute()
          case unknownOperation =>
            throw HiveAcidErrors.invalidOperationType(unknownOperation.toString)
        }

      logInfo(s"Write Operation being performed to table ${hiveAcidMetadata.tableName}: $operationType")

      val touchedPartitions = sparkSession.sparkContext.runJob(
        resultRDD, processRddPartition
      ).flatten.toSet

      // Add new partition to table metadata under the transaction.
      val existingPartitions = hiveAcidMetadata.getRawPartitions()
        .map(_.getSpec)
        .map(_.asScala.toMap)

      val newPartitions = touchedPartitions -- existingPartitions

      logDebug(s"existing partitions: ${touchedPartitions.size}, " +
        s"partitions touched: ${touchedPartitions.size}, " +
        s"new partitions to add to metastore: ${newPartitions.size}")

      if (newPartitions.nonEmpty) {
        AlterTableAddPartitionCommand(
          new TableIdentifier(hiveAcidMetadata.tableName, Option(hiveAcidMetadata.dbName)),
          newPartitions.toSeq.map(p => (p, None)),
          ifNotExists = true).run(sparkSession)
        // FIXME: Add the notification events for replication et al.
        //
        logDebug("new partitions added successfully")
      }
      if (touchedPartitions.size > 0) {
        val touchedPartitionNames = touchedPartitions.map (
          PartitioningUtils.getPathFragment(_, partitionColumns))

        curTxn.addDynamicPartitions(curSnapshot.currentWriteId, hiveAcidMetadata.dbName,
          hiveAcidMetadata.tableName, operationType, touchedPartitionNames)
      }

    } catch {
      case e: Exception =>
        logError("Exception", e)
        throw e
    }
  }
}
